"""Config flow for Locative."""
from openpeerpower.helpers import config_entry_flow

from .const import DOMAIN

config_entry_flow.register_webhook_flow(
    DOMAIN,
    "Locative Webhook",
    {"docs_url": "https://www.openpeerpower.io/integrations/locative/"},
)



"""Helper methods to help with platform discovery.

There are two different types of discoveries that can be fired/listened for.
 - listen/discover is for services. These are targeted at a component.
 - listen_platform/discover_platform is for platforms. These are used by
   components to allow discovery of their platforms.
"""
from typing import Any, Callable, Dict, Optional, TypedDict

from openpeerpower import core, setup
from openpeerpower.core import CALLBACK_TYPE
from openpeerpower.loader import bind_opp

from .dispatcher import async_dispatcher_connect, async_dispatcher_send
from .typing import ConfigType, DiscoveryInfoType

SIGNAL_PLATFORM_DISCOVERED = "discovery.platform_discovered_{}"
EVENT_LOAD_PLATFORM = "load_platform.{}"
ATTR_PLATFORM = "platform"
ATTR_DISCOVERED = "discovered"

# mypy: disallow-any-generics


class DiscoveryDict(TypedDict):
    """Discovery data."""

    service: str
    platform: Optional[str]
    discovered: Optional[DiscoveryInfoType]


@core.callback
@bind_opp
def async_listen(
    opp: core.OpenPeerPower,
    service: str,
    callback: CALLBACK_TYPE,
) -> None:
    """Set up listener for discovery of specific service.

    Service can be a string or a list/tuple.
    """
    job = core.OppJob(callback)

    async def discovery_event_listener(discovered: DiscoveryDict) -> None:
        """Listen for discovery events."""
        task = opp.async_run_opp_job(
            job, discovered["service"], discovered["discovered"]
        )
        if task:
            await task

    async_dispatcher_connect(
        opp, SIGNAL_PLATFORM_DISCOVERED.format(service), discovery_event_listener
    )


@bind_opp
def discover(
    opp: core.OpenPeerPower,
    service: str,
    discovered: DiscoveryInfoType,
    component: str,
    opp_config: ConfigType,
) -> None:
    """Fire discovery event. Can ensure a component is loaded."""
    opp.add_job(
        async_discover(opp, service, discovered, component, opp_config)  # type: ignore
    )


@bind_opp
async def async_discover(
    opp: core.OpenPeerPower,
    service: str,
    discovered: Optional[DiscoveryInfoType],
    component: Optional[str],
    opp_config: ConfigType,
) -> None:
    """Fire discovery event. Can ensure a component is loaded."""
    if component is not None and component not in opp.config.components:
        await setup.async_setup_component(opp, component, opp_config)

    data: DiscoveryDict = {
        "service": service,
        "platform": None,
        "discovered": discovered,
    }

    async_dispatcher_send(opp, SIGNAL_PLATFORM_DISCOVERED.format(service), data)


@bind_opp
def async_listen_platform(
    opp: core.OpenPeerPower,
    component: str,
    callback: Callable[[str, Optional[Dict[str, Any]]], Any],
) -> None:
    """Register a platform loader listener.

    This method must be run in the event loop.
    """
    service = EVENT_LOAD_PLATFORM.format(component)
    job = core.OppJob(callback)

    async def discovery_platform_listener(discovered: DiscoveryDict) -> None:
        """Listen for platform discovery events."""
        platform = discovered["platform"]

        if not platform:
            return

        task = opp.async_run_opp_job(job, platform, discovered.get("discovered"))
        if task:
            await task

    async_dispatcher_connect(
        opp, SIGNAL_PLATFORM_DISCOVERED.format(service), discovery_platform_listener
    )


@bind_opp
def load_platform(
    opp: core.OpenPeerPower,
    component: str,
    platform: str,
    discovered: DiscoveryInfoType,
    opp_config: ConfigType,
) -> None:
    """Load a component and platform dynamically."""
    opp.add_job(
        async_load_platform(  # type: ignore
            opp, component, platform, discovered, opp_config
        )
    )


@bind_opp
async def async_load_platform(
    opp: core.OpenPeerPower,
    component: str,
    platform: str,
    discovered: DiscoveryInfoType,
    opp_config: ConfigType,
) -> None:
    """Load a component and platform dynamically.

    Use `async_listen_platform` to register a callback for these events.

    Warning: Do not await this inside a setup method to avoid a dead lock.
    Use `opp.async_create_task(async_load_platform(..))` instead.
    """
    assert opp_config, "You need to pass in the real opp config"

    setup_success = True

    if component not in opp.config.components:
        setup_success = await setup.async_setup_component(opp, component, opp_config)

    # No need to send signal if we could not set up component
    if not setup_success:
        return

    service = EVENT_LOAD_PLATFORM.format(component)

    data: DiscoveryDict = {
        "service": service,
        "platform": platform,
        "discovered": discovered,
    }

    async_dispatcher_send(opp, SIGNAL_PLATFORM_DISCOVERED.format(service), data)



"""Run Open Peer Power."""
import asyncio
from concurrent.futures import ThreadPoolExecutor
import dataclasses
import logging
from typing import Any, Dict, Optional

from openpeerpower import bootstrap
from openpeerpower.core import callback
from openpeerpower.helpers.frame import warn_use

# mypy: disallow-any-generics

#
# Python 3.8 has significantly less workers by default
# than Python 3.7.  In order to be consistent between
# supported versions, we need to set max_workers.
#
# In most cases the workers are not I/O bound, as they
# are sleeping/blocking waiting for data from integrations
# updating so this number should be higher than the default
# use case.
#
MAX_EXECUTOR_WORKERS = 64


@dataclasses.dataclass
class RuntimeConfig:
    """Class to hold the information for running Open Peer Power."""

    config_dir: str
    skip_pip: bool = False
    safe_mode: bool = False

    verbose: bool = False

    log_rotate_days: Optional[int] = None
    log_file: Optional[str] = None
    log_no_color: bool = False

    debug: bool = False
    open_ui: bool = False


class OppEventLoopPolicy(asyncio.DefaultEventLoopPolicy):  # type: ignore[valid-type,misc]
    """Event loop policy for Open Peer Power."""

    def __init__(self, debug: bool) -> None:
        """Init the event loop policy."""
        super().__init__()
        self.debug = debug

    @property
    def loop_name(self) -> str:
        """Return name of the loop."""
        return self._loop_factory.__name__  # type: ignore

    def new_event_loop(self) -> asyncio.AbstractEventLoop:
        """Get the event loop."""
        loop: asyncio.AbstractEventLoop = super().new_event_loop()
        loop.set_exception_handler(_async_loop_exception_handler)
        if self.debug:
            loop.set_debug(True)

        executor = ThreadPoolExecutor(
            thread_name_prefix="SyncWorker", max_workers=MAX_EXECUTOR_WORKERS
        )
        loop.set_default_executor(executor)
        loop.set_default_executor = warn_use(  # type: ignore
            loop.set_default_executor, "sets default executor on the event loop"
        )

        # Shut down executor when we shut down loop
        orig_close = loop.close

        def close() -> None:
            executor.shutdown(wait=True)
            orig_close()

        loop.close = close  # type: ignore

        return loop


@callback
def _async_loop_exception_handler(_: Any, context: Dict[str, Any]) -> None:
    """Handle all exception inside the core loop."""
    kwargs = {}
    exception = context.get("exception")
    if exception:
        kwargs["exc_info"] = (type(exception), exception, exception.__traceback__)

    logging.getLogger(__package__).error(
        "Error doing job: %s", context["message"], **kwargs  # type: ignore
    )


async def setup_and_run_opp(runtime_config: RuntimeConfig) -> int:
    """Set up Open Peer Power and run."""
    opp = await bootstrap.async_setup_opp(runtime_config)

    if opp is None:
        return 1

    return await opp.async_run()


def run(runtime_config: RuntimeConfig) -> int:
    """Run Open Peer Power."""
    asyncio.set_event_loop_policy(OppEventLoopPolicy(runtime_config.debug))
    return asyncio.run(setup_and_run_opp(runtime_config))



"""Discord platform for notify component."""
import logging
import os.path

import discord
import voluptuous as vol

from openpeerpower.components.notify import (
    ATTR_DATA,
    ATTR_TARGET,
    PLATFORM_SCHEMA,
    BaseNotificationService,
)
from openpeerpower.const import CONF_TOKEN
import openpeerpower.helpers.config_validation as cv

_LOGGER = logging.getLogger(__name__)

ATTR_EMBED = "embed"
ATTR_EMBED_AUTHOR = "author"
ATTR_EMBED_FIELDS = "fields"
ATTR_EMBED_FOOTER = "footer"
ATTR_EMBED_THUMBNAIL = "thumbnail"
ATTR_IMAGES = "images"

PLATFORM_SCHEMA = PLATFORM_SCHEMA.extend({vol.Required(CONF_TOKEN): cv.string})


def get_service(opp, config, discovery_info=None):
    """Get the Discord notification service."""
    token = config[CONF_TOKEN]
    return DiscordNotificationService(opp, token)


class DiscordNotificationService(BaseNotificationService):
    """Implement the notification service for Discord."""

    def __init__(self, opp, token):
        """Initialize the service."""
        self.token = token
        self.opp = opp

    def file_exists(self, filename):
        """Check if a file exists on disk and is in authorized path."""
        if not self.opp.config.is_allowed_path(filename):
            return False
        return os.path.isfile(filename)

    async def async_send_message(self, message, **kwargs):
        """Login to Discord, send message to channel(s) and log out."""
        discord.VoiceClient.warn_nacl = False
        discord_bot = discord.Client()
        images = None
        embedding = None

        if ATTR_TARGET not in kwargs:
            _LOGGER.error("No target specified")
            return None

        data = kwargs.get(ATTR_DATA) or {}

        if ATTR_EMBED in data:
            embedding = data[ATTR_EMBED]
            fields = embedding.get(ATTR_EMBED_FIELDS)

        if ATTR_IMAGES in data:
            images = []

            for image in data.get(ATTR_IMAGES):
                image_exists = await self.opp.async_add_executor_job(
                    self.file_exists, image
                )

                if image_exists:
                    images.append(image)
                else:
                    _LOGGER.warning("Image not found: %s", image)

        # pylint: disable=unused-variable
        @discord_bot.event
        async def on_ready():
            """Send the messages when the bot is ready."""
            try:
                for channelid in kwargs[ATTR_TARGET]:
                    channelid = int(channelid)
                    channel = discord_bot.get_channel(
                        channelid
                    ) or discord_bot.get_user(channelid)

                    if channel is None:
                        _LOGGER.warning("Channel not found for ID: %s", channelid)
                        continue
                    # Must create new instances of File for each channel.
                    files = None
                    if images:
                        files = []
                        for image in images:
                            files.append(discord.File(image))
                    if embedding:
                        embed = discord.Embed(**embedding)
                        if fields:
                            for field_num, field_name in enumerate(fields):
                                embed.add_field(**fields[field_num])
                        if ATTR_EMBED_FOOTER in embedding:
                            embed.set_footer(**embedding[ATTR_EMBED_FOOTER])
                        if ATTR_EMBED_AUTHOR in embedding:
                            embed.set_author(**embedding[ATTR_EMBED_AUTHOR])
                        if ATTR_EMBED_THUMBNAIL in embedding:
                            embed.set_thumbnail(**embedding[ATTR_EMBED_THUMBNAIL])
                        await channel.send(message, files=files, embed=embed)
                    else:
                        await channel.send(message, files=files)
            except (discord.errors.HTTPException, discord.errors.NotFound) as error:
                _LOGGER.warning("Communication error: %s", error)
            await discord_bot.logout()
            await discord_bot.close()

        # Using reconnect=False prevents multiple ready events to be fired.
        await discord_bot.start(self.token, reconnect=False)



"""Describe group states."""


from openpeerpower.components.group import GroupIntegrationRegistry
from openpeerpower.const import STATE_OFF
from openpeerpower.core import callback
from openpeerpower.helpers.typing import OpenPeerPowerType

from .const import HVAC_MODE_OFF, HVAC_MODES


@callback
def async_describe_on_off_states(
    opp: OpenPeerPowerType, registry: GroupIntegrationRegistry
) -> None:
    """Describe group on off states."""
    registry.on_off_states(
        set(HVAC_MODES) - {HVAC_MODE_OFF},
        STATE_OFF,
    )



"""The Ondilo ICO integration."""
import asyncio

from openpeerpower.config_entries import ConfigEntry
from openpeerpower.core import OpenPeerPower
from openpeerpower.helpers import config_entry_oauth2_flow

from . import api, config_flow
from .const import DOMAIN
from .oauth_impl import OndiloOauth2Implementation

PLATFORMS = ["sensor"]


async def async_setup(opp: OpenPeerPower, config: dict):
    """Set up the Ondilo ICO component."""
    opp.data.setdefault(DOMAIN, {})

    return True


async def async_setup_entry(opp: OpenPeerPower, entry: ConfigEntry) -> bool:
    """Set up Ondilo ICO from a config entry."""

    config_flow.OAuth2FlowHandler.async_register_implementation(
        opp,
        OndiloOauth2Implementation(opp),
    )

    implementation = (
        await config_entry_oauth2_flow.async_get_config_entry_implementation(opp, entry)
    )

    opp.data[DOMAIN][entry.entry_id] = api.OndiloClient(opp, entry, implementation)

    for platform in PLATFORMS:
        opp.async_create_task(
            opp.config_entries.async_forward_entry_setup(entry, platform)
        )

    return True


async def async_unload_entry(opp: OpenPeerPower, entry: ConfigEntry):
    """Unload a config entry."""
    unload_ok = all(
        await asyncio.gather(
            *[
                opp.config_entries.async_forward_entry_unload(entry, platform)
                for platform in PLATFORMS
            ]
        )
    )
    if unload_ok:
        opp.data[DOMAIN].pop(entry.entry_id)

    return unload_ok



"""Errors for the Media Player component."""
from openpeerpower.exceptions import OpenPeerPowerError


class MediaPlayerException(OpenPeerPowerError):
    """Base class for Media Player exceptions."""


class BrowseError(MediaPlayerException):
    """Error while browsing."""



"""Describe group states."""


from openpeerpower.components.group import GroupIntegrationRegistry
from openpeerpower.const import STATE_CLOSED, STATE_OPEN
from openpeerpower.core import callback
from openpeerpower.helpers.typing import OpenPeerPowerType


@callback
def async_describe_on_off_states(
    opp: OpenPeerPowerType, registry: GroupIntegrationRegistry
) -> None:
    """Describe group on off states."""
    # On means open, Off means closed
    registry.on_off_states({STATE_OPEN}, STATE_CLOSED)



"""Provides device automations for Lock."""
from typing import List, Optional

import voluptuous as vol

from openpeerpower.const import (
    ATTR_ENTITY_ID,
    ATTR_SUPPORTED_FEATURES,
    CONF_DEVICE_ID,
    CONF_DOMAIN,
    CONF_ENTITY_ID,
    CONF_TYPE,
    SERVICE_LOCK,
    SERVICE_OPEN,
    SERVICE_UNLOCK,
)
from openpeerpower.core import Context, OpenPeerPower
from openpeerpower.helpers import entity_registry
import openpeerpower.helpers.config_validation as cv

from . import DOMAIN, SUPPORT_OPEN

ACTION_TYPES = {"lock", "unlock", "open"}

ACTION_SCHEMA = cv.DEVICE_ACTION_BASE_SCHEMA.extend(
    {
        vol.Required(CONF_TYPE): vol.In(ACTION_TYPES),
        vol.Required(CONF_ENTITY_ID): cv.entity_domain(DOMAIN),
    }
)


async def async_get_actions(opp: OpenPeerPower, device_id: str) -> List[dict]:
    """List device actions for Lock devices."""
    registry = await entity_registry.async_get_registry(opp)
    actions = []

    # Get all the integrations entities for this device
    for entry in entity_registry.async_entries_for_device(registry, device_id):
        if entry.domain != DOMAIN:
            continue

        # Add actions for each entity that belongs to this integration
        actions.append(
            {
                CONF_DEVICE_ID: device_id,
                CONF_DOMAIN: DOMAIN,
                CONF_ENTITY_ID: entry.entity_id,
                CONF_TYPE: "lock",
            }
        )
        actions.append(
            {
                CONF_DEVICE_ID: device_id,
                CONF_DOMAIN: DOMAIN,
                CONF_ENTITY_ID: entry.entity_id,
                CONF_TYPE: "unlock",
            }
        )

        state = opp.states.get(entry.entity_id)
        if state:
            features = state.attributes.get(ATTR_SUPPORTED_FEATURES, 0)
            if features & (SUPPORT_OPEN):
                actions.append(
                    {
                        CONF_DEVICE_ID: device_id,
                        CONF_DOMAIN: DOMAIN,
                        CONF_ENTITY_ID: entry.entity_id,
                        CONF_TYPE: "open",
                    }
                )

    return actions


async def async_call_action_from_config(
    opp: OpenPeerPower, config: dict, variables: dict, context: Optional[Context]
) -> None:
    """Execute a device action."""
    config = ACTION_SCHEMA(config)

    service_data = {ATTR_ENTITY_ID: config[CONF_ENTITY_ID]}

    if config[CONF_TYPE] == "lock":
        service = SERVICE_LOCK
    elif config[CONF_TYPE] == "unlock":
        service = SERVICE_UNLOCK
    elif config[CONF_TYPE] == "open":
        service = SERVICE_OPEN

    await opp.services.async_call(
        DOMAIN, service, service_data, blocking=True, context=context
    )



"""Platform for the iZone AC."""
import voluptuous as vol

from openpeerpower import config_entries
from openpeerpower.const import CONF_EXCLUDE
import openpeerpower.helpers.config_validation as cv
from openpeerpower.helpers.typing import ConfigType, OpenPeerPowerType

from .const import DATA_CONFIG, IZONE
from .discovery import async_start_discovery_service, async_stop_discovery_service

CONFIG_SCHEMA = vol.Schema(
    {
        IZONE: vol.Schema(
            {
                vol.Optional(CONF_EXCLUDE, default=[]): vol.All(
                    cv.ensure_list, [cv.string]
                )
            }
        )
    },
    extra=vol.ALLOW_EXTRA,
)


async def async_setup(opp: OpenPeerPowerType, config: ConfigType):
    """Register the iZone component config."""
    conf = config.get(IZONE)
    if not conf:
        return True

    opp.data[DATA_CONFIG] = conf

    # Explicitly added in the config file, create a config entry.
    opp.async_create_task(
        opp.config_entries.flow.async_init(
            IZONE, context={"source": config_entries.SOURCE_IMPORT}
        )
    )

    return True


async def async_setup_entry(opp, entry):
    """Set up from a config entry."""
    await async_start_discovery_service(opp)

    opp.async_create_task(
        opp.config_entries.async_forward_entry_setup(entry, "climate")
    )
    return True


async def async_unload_entry(opp, entry):
    """Unload the config entry and stop discovery process."""
    await async_stop_discovery_service(opp)
    await opp.config_entries.async_forward_entry_unload(entry, "climate")
    return True



"""Utility methods for the Insteon platform."""
from pyinsteon.device_types import (
    ClimateControl_Thermostat,
    ClimateControl_WirelessThermostat,
    DimmableLightingControl,
    DimmableLightingControl_DinRail,
    DimmableLightingControl_FanLinc,
    DimmableLightingControl_InLineLinc,
    DimmableLightingControl_KeypadLinc_6,
    DimmableLightingControl_KeypadLinc_8,
    DimmableLightingControl_LampLinc,
    DimmableLightingControl_OutletLinc,
    DimmableLightingControl_SwitchLinc,
    DimmableLightingControl_ToggleLinc,
    GeneralController_ControlLinc,
    GeneralController_MiniRemote_4,
    GeneralController_MiniRemote_8,
    GeneralController_MiniRemote_Switch,
    GeneralController_RemoteLinc,
    SecurityHealthSafety_DoorSensor,
    SecurityHealthSafety_LeakSensor,
    SecurityHealthSafety_MotionSensor,
    SecurityHealthSafety_OpenCloseSensor,
    SecurityHealthSafety_Smokebridge,
    SensorsActuators_IOLink,
    SwitchedLightingControl,
    SwitchedLightingControl_ApplianceLinc,
    SwitchedLightingControl_DinRail,
    SwitchedLightingControl_InLineLinc,
    SwitchedLightingControl_KeypadLinc_6,
    SwitchedLightingControl_KeypadLinc_8,
    SwitchedLightingControl_OnOffOutlet,
    SwitchedLightingControl_OutletLinc,
    SwitchedLightingControl_SwitchLinc,
    SwitchedLightingControl_ToggleLinc,
    WindowCovering,
    X10Dimmable,
    X10OnOff,
    X10OnOffSensor,
)

from openpeerpower.components.binary_sensor import DOMAIN as BINARY_SENSOR
from openpeerpower.components.climate import DOMAIN as CLIMATE
from openpeerpower.components.cover import DOMAIN as COVER
from openpeerpower.components.fan import DOMAIN as FAN
from openpeerpower.components.light import DOMAIN as LIGHT
from openpeerpower.components.switch import DOMAIN as SWITCH

from .const import ON_OFF_EVENTS

DEVICE_PLATFORM = {
    DimmableLightingControl: {LIGHT: [1], ON_OFF_EVENTS: [1]},
    DimmableLightingControl_DinRail: {LIGHT: [1], ON_OFF_EVENTS: [1]},
    DimmableLightingControl_FanLinc: {LIGHT: [1], FAN: [2], ON_OFF_EVENTS: [1, 2]},
    DimmableLightingControl_InLineLinc: {LIGHT: [1], ON_OFF_EVENTS: [1]},
    DimmableLightingControl_KeypadLinc_6: {
        LIGHT: [1],
        SWITCH: [3, 4, 5, 6],
        ON_OFF_EVENTS: [1, 3, 4, 5, 6],
    },
    DimmableLightingControl_KeypadLinc_8: {
        LIGHT: [1],
        SWITCH: range(2, 9),
        ON_OFF_EVENTS: range(1, 9),
    },
    DimmableLightingControl_LampLinc: {LIGHT: [1], ON_OFF_EVENTS: [1]},
    DimmableLightingControl_OutletLinc: {LIGHT: [1], ON_OFF_EVENTS: [1]},
    DimmableLightingControl_SwitchLinc: {LIGHT: [1], ON_OFF_EVENTS: [1]},
    DimmableLightingControl_ToggleLinc: {LIGHT: [1], ON_OFF_EVENTS: [1]},
    GeneralController_ControlLinc: {ON_OFF_EVENTS: [1]},
    GeneralController_MiniRemote_4: {ON_OFF_EVENTS: range(1, 5)},
    GeneralController_MiniRemote_8: {ON_OFF_EVENTS: range(1, 9)},
    GeneralController_MiniRemote_Switch: {ON_OFF_EVENTS: [1, 2]},
    GeneralController_RemoteLinc: {ON_OFF_EVENTS: [1]},
    SecurityHealthSafety_DoorSensor: {BINARY_SENSOR: [1, 3, 4], ON_OFF_EVENTS: [1]},
    SecurityHealthSafety_LeakSensor: {BINARY_SENSOR: [2, 4]},
    SecurityHealthSafety_MotionSensor: {BINARY_SENSOR: [1, 2, 3], ON_OFF_EVENTS: [1]},
    SecurityHealthSafety_OpenCloseSensor: {BINARY_SENSOR: [1]},
    SecurityHealthSafety_Smokebridge: {BINARY_SENSOR: [1, 2, 3, 4, 6, 7]},
    SensorsActuators_IOLink: {SWITCH: [1], BINARY_SENSOR: [2], ON_OFF_EVENTS: [1, 2]},
    SwitchedLightingControl: {SWITCH: [1], ON_OFF_EVENTS: [1]},
    SwitchedLightingControl_ApplianceLinc: {SWITCH: [1], ON_OFF_EVENTS: [1]},
    SwitchedLightingControl_DinRail: {SWITCH: [1], ON_OFF_EVENTS: [1]},
    SwitchedLightingControl_InLineLinc: {SWITCH: [1], ON_OFF_EVENTS: [1]},
    SwitchedLightingControl_KeypadLinc_6: {
        SWITCH: [1, 3, 4, 5, 6],
        ON_OFF_EVENTS: [1, 3, 4, 5, 6],
    },
    SwitchedLightingControl_KeypadLinc_8: {
        SWITCH: range(1, 9),
        ON_OFF_EVENTS: range(1, 9),
    },
    SwitchedLightingControl_OnOffOutlet: {SWITCH: [1, 2], ON_OFF_EVENTS: [1, 2]},
    SwitchedLightingControl_OutletLinc: {SWITCH: [1], ON_OFF_EVENTS: [1]},
    SwitchedLightingControl_SwitchLinc: {SWITCH: [1], ON_OFF_EVENTS: [1]},
    SwitchedLightingControl_ToggleLinc: {SWITCH: [1], ON_OFF_EVENTS: [1]},
    ClimateControl_Thermostat: {CLIMATE: [1]},
    ClimateControl_WirelessThermostat: {CLIMATE: [1]},
    WindowCovering: {COVER: [1]},
    X10Dimmable: {LIGHT: [1]},
    X10OnOff: {SWITCH: [1]},
    X10OnOffSensor: {BINARY_SENSOR: [1]},
}


def get_device_platforms(device):
    """Return the OP platforms for a device type."""
    return DEVICE_PLATFORM.get(type(device), {}).keys()


def get_platform_groups(device, domain) -> dict:
    """Return the platforms that a device belongs in."""
    return DEVICE_PLATFORM.get(type(device), {}).get(domain, {})



"""Support for Gogogate2 garage Doors."""
from typing import Callable, List, Optional

from gogogate2_api.common import get_configured_doors

from openpeerpower.config_entries import ConfigEntry
from openpeerpower.const import DEVICE_CLASS_BATTERY
from openpeerpower.core import OpenPeerPower
from openpeerpower.helpers.entity import Entity

from .common import GoGoGate2Entity, get_data_update_coordinator

SENSOR_ID_WIRED = "WIRE"


async def async_setup_entry(
    opp: OpenPeerPower,
    config_entry: ConfigEntry,
    async_add_entities: Callable[[List[Entity], Optional[bool]], None],
) -> None:
    """Set up the config entry."""
    data_update_coordinator = get_data_update_coordinator(opp, config_entry)

    async_add_entities(
        [
            DoorSensor(config_entry, data_update_coordinator, door)
            for door in get_configured_doors(data_update_coordinator.data)
            if door.sensorid and door.sensorid != SENSOR_ID_WIRED
        ]
    )


class DoorSensor(GoGoGate2Entity):
    """Sensor entity for goggate2."""

    @property
    def name(self):
        """Return the name of the door."""
        return f"{self._get_door().name} battery"

    @property
    def device_class(self):
        """Return the class of this device, from component DEVICE_CLASSES."""
        return DEVICE_CLASS_BATTERY

    @property
    def state(self):
        """Return the state of the entity."""
        door = self._get_door()
        return door.voltage  # This is a percentage, not an absolute voltage

    @property
    def device_state_attributes(self):
        """Return the state attributes."""
        door = self._get_door()
        if door.sensorid is not None:
            return {"door_id": door.door_id, "sensor_id": door.sensorid}
        return None



"""Describe group states."""


from openpeerpower.components.group import GroupIntegrationRegistry
from openpeerpower.const import (
    STATE_IDLE,
    STATE_OFF,
    STATE_ON,
    STATE_PAUSED,
    STATE_PLAYING,
)
from openpeerpower.core import callback
from openpeerpower.helpers.typing import OpenPeerPowerType


@callback
def async_describe_on_off_states(
    opp: OpenPeerPowerType, registry: GroupIntegrationRegistry
) -> None:
    """Describe group on off states."""
    registry.on_off_states(
        {STATE_ON, STATE_PAUSED, STATE_PLAYING, STATE_IDLE}, STATE_OFF
    )



"""Implement the auth feature from Opp.io for Add-ons."""
from ipaddress import ip_address
import logging
import os

from aiohttp import web
from aiohttp.web_exceptions import HTTPNotFound, HTTPUnauthorized
import voluptuous as vol

from openpeerpower.auth.models import User
from openpeerpower.auth.providers import openpeerpower as auth_op
from openpeerpower.components.http import OpenPeerPowerView
from openpeerpower.components.http.const import KEY_OPP_USER
from openpeerpower.components.http.data_validator import RequestDataValidator
from openpeerpower.const import HTTP_OK
from openpeerpower.core import callback
import openpeerpower.helpers.config_validation as cv
from openpeerpower.helpers.typing import OpenPeerPowerType

from .const import ATTR_ADDON, ATTR_PASSWORD, ATTR_USERNAME

_LOGGER = logging.getLogger(__name__)


@callback
def async_setup_auth_view(opp: OpenPeerPowerType, user: User):
    """Auth setup."""
    oppio_auth = OppIOAuth(opp, user)
    oppio_password_reset = OppIOPasswordReset(opp, user)

    opp.http.register_view(oppio_auth)
    opp.http.register_view(oppio_password_reset)


class OppIOBaseAuth(OpenPeerPowerView):
    """Opp.io view to handle auth requests."""

    def __init__(self, opp: OpenPeerPowerType, user: User):
        """Initialize WebView."""
        self.opp = opp
        self.user = user

    def _check_access(self, request: web.Request):
        """Check if this call is from Supervisor."""
        # Check caller IP
        oppio_ip = os.environ["OPPIO"].split(":")[0]
        if ip_address(request.transport.get_extra_info("peername")[0]) != ip_address(
            oppio_ip
        ):
            _LOGGER.error("Invalid auth request from %s", request.remote)
            raise HTTPUnauthorized()

        # Check caller token
        if request[KEY_OPP_USER].id != self.user.id:
            _LOGGER.error("Invalid auth request from %s", request[KEY_OPP_USER].name)
            raise HTTPUnauthorized()


class OppIOAuth(OppIOBaseAuth):
    """Opp.io view to handle auth requests."""

    name = "api:oppio:auth"
    url = "/api/oppio_auth"

    @RequestDataValidator(
        vol.Schema(
            {
                vol.Required(ATTR_USERNAME): cv.string,
                vol.Required(ATTR_PASSWORD): cv.string,
                vol.Required(ATTR_ADDON): cv.string,
            },
            extra=vol.ALLOW_EXTRA,
        )
    )
    async def post(self, request, data):
        """Handle auth requests."""
        self._check_access(request)
        provider = auth_op.async_get_provider(request.app["opp"])

        try:
            await provider.async_validate_login(
                data[ATTR_USERNAME], data[ATTR_PASSWORD]
            )
        except auth_op.InvalidAuth:
            raise HTTPNotFound() from None

        return web.Response(status=HTTP_OK)


class OppIOPasswordReset(OppIOBaseAuth):
    """Opp.io view to handle password reset requests."""

    name = "api:oppio:auth:password:reset"
    url = "/api/oppio_auth/password_reset"

    @RequestDataValidator(
        vol.Schema(
            {
                vol.Required(ATTR_USERNAME): cv.string,
                vol.Required(ATTR_PASSWORD): cv.string,
            },
            extra=vol.ALLOW_EXTRA,
        )
    )
    async def post(self, request, data):
        """Handle password reset requests."""
        self._check_access(request)
        provider = auth_op.async_get_provider(request.app["opp"])

        try:
            await provider.async_change_password(
                data[ATTR_USERNAME], data[ATTR_PASSWORD]
            )
        except auth_op.InvalidUser as err:
            raise HTTPNotFound() from err

        return web.Response(status=HTTP_OK)



"""Config flow for kmtronic integration."""
import logging

import aiohttp
from pykmtronic.auth import Auth
from pykmtronic.hub import KMTronicHubAPI
import voluptuous as vol

from openpeerpower import config_entries, core, exceptions
from openpeerpower.helpers import aiohttp_client

from .const import CONF_HOSTNAME, CONF_PASSWORD, CONF_USERNAME
from .const import DOMAIN  # pylint:disable=unused-import

_LOGGER = logging.getLogger(__name__)

DATA_SCHEMA = vol.Schema({CONF_HOSTNAME: str, CONF_USERNAME: str, CONF_PASSWORD: str})


async def validate_input(opp: core.OpenPeerPower, data):
    """Validate the user input allows us to connect."""

    session = aiohttp_client.async_get_clientsession(opp)
    auth = Auth(
        session,
        f"http://{data[CONF_HOSTNAME]}",
        data[CONF_USERNAME],
        data[CONF_PASSWORD],
    )
    hub = KMTronicHubAPI(auth)

    try:
        await hub.async_get_status()
    except aiohttp.client_exceptions.ClientResponseError as err:
        raise InvalidAuth from err
    except aiohttp.client_exceptions.ClientConnectorError as err:
        raise CannotConnect from err

    return data


class ConfigFlow(config_entries.ConfigFlow, domain=DOMAIN):
    """Handle a config flow for kmtronic."""

    VERSION = 1
    CONNECTION_CLASS = config_entries.CONN_CLASS_LOCAL_POLL

    async def async_step_user(self, user_input=None):
        """Handle the initial step."""
        errors = {}
        if user_input is not None:
            try:
                info = await validate_input(self.opp, user_input)

                return self.async_create_entry(title=info["host"], data=user_input)
            except CannotConnect:
                errors["base"] = "cannot_connect"
            except InvalidAuth:
                errors["base"] = "invalid_auth"
            except Exception:  # pylint: disable=broad-except
                _LOGGER.exception("Unexpected exception")
                errors["base"] = "unknown"

        return self.async_show_form(
            step_id="user", data_schema=DATA_SCHEMA, errors=errors
        )


class CannotConnect(exceptions.OpenPeerPowerError):
    """Error to indicate we cannot connect."""


class InvalidAuth(exceptions.OpenPeerPowerError):
    """Error to indicate there is invalid auth."""



"""Config validation helper for the script integration."""
import asyncio

import voluptuous as vol

from openpeerpower.config import async_log_exception
from openpeerpower.const import CONF_SEQUENCE
from openpeerpower.exceptions import OpenPeerPowerError
from openpeerpower.helpers import config_validation as cv
from openpeerpower.helpers.script import async_validate_action_config

from . import DOMAIN, SCRIPT_ENTRY_SCHEMA


async def async_validate_config_item(opp, config, full_config=None):
    """Validate config item."""
    config = SCRIPT_ENTRY_SCHEMA(config)
    config[CONF_SEQUENCE] = await asyncio.gather(
        *[async_validate_action_config(opp, action) for action in config[CONF_SEQUENCE]]
    )

    return config


async def _try_async_validate_config_item(opp, object_id, config, full_config=None):
    """Validate config item."""
    try:
        cv.slug(object_id)
        config = await async_validate_config_item(opp, config, full_config)
    except (vol.Invalid, OpenPeerPowerError) as ex:
        async_log_exception(ex, DOMAIN, full_config or config, opp)
        return None

    return config


async def async_validate_config(opp, config):
    """Validate config."""
    if DOMAIN in config:
        validated_config = {}
        for object_id, cfg in config[DOMAIN].items():
            cfg = await _try_async_validate_config_item(opp, object_id, cfg, config)
            if cfg is not None:
                validated_config[object_id] = cfg
        config[DOMAIN] = validated_config

    return config



"""The Litter-Robot integration."""
import asyncio

from pylitterbot.exceptions import LitterRobotException, LitterRobotLoginException

from openpeerpower.config_entries import ConfigEntry
from openpeerpower.core import OpenPeerPower
from openpeerpower.exceptions import ConfigEntryNotReady

from .const import DOMAIN
from .hub import LitterRobotHub

PLATFORMS = ["sensor", "switch", "vacuum"]


async def async_setup(opp: OpenPeerPower, config: dict):
    """Set up the Litter-Robot component."""
    opp.data.setdefault(DOMAIN, {})

    return True


async def async_setup_entry(opp: OpenPeerPower, entry: ConfigEntry):
    """Set up Litter-Robot from a config entry."""
    hub = opp.data[DOMAIN][entry.entry_id] = LitterRobotHub(opp, entry.data)
    try:
        await hub.login(load_robots=True)
    except LitterRobotLoginException:
        return False
    except LitterRobotException as ex:
        raise ConfigEntryNotReady from ex

    for platform in PLATFORMS:
        opp.async_create_task(
            opp.config_entries.async_forward_entry_setup(entry, platform)
        )

    return True


async def async_unload_entry(opp: OpenPeerPower, entry: ConfigEntry):
    """Unload a config entry."""
    unload_ok = all(
        await asyncio.gather(
            *[
                opp.config_entries.async_forward_entry_unload(entry, platform)
                for platform in PLATFORMS
            ]
        )
    )
    if unload_ok:
        opp.data[DOMAIN].pop(entry.entry_id)

    return unload_ok



"""Support to interface with universal remote control devices."""
from datetime import timedelta
import functools as ft
import logging
from typing import Any, Dict, Iterable, List, Optional, cast

import voluptuous as vol

from openpeerpower.config_entries import ConfigEntry
from openpeerpower.const import (
    ATTR_COMMAND,
    SERVICE_TOGGLE,
    SERVICE_TURN_OFF,
    SERVICE_TURN_ON,
    STATE_ON,
)
import openpeerpower.helpers.config_validation as cv
from openpeerpower.helpers.config_validation import (  # noqa: F401
    PLATFORM_SCHEMA,
    PLATFORM_SCHEMA_BASE,
    make_entity_service_schema,
)
from openpeerpower.helpers.entity import ToggleEntity
from openpeerpower.helpers.entity_component import EntityComponent
from openpeerpower.helpers.typing import ConfigType, OpenPeerPowerType
from openpeerpower.loader import bind_opp

# mypy: allow-untyped-calls, allow-untyped-defs, no-check-untyped-defs

_LOGGER = logging.getLogger(__name__)

ATTR_ACTIVITY = "activity"
ATTR_ACTIVITY_LIST = "activity_list"
ATTR_CURRENT_ACTIVITY = "current_activity"
ATTR_COMMAND_TYPE = "command_type"
ATTR_DEVICE = "device"
ATTR_NUM_REPEATS = "num_repeats"
ATTR_DELAY_SECS = "delay_secs"
ATTR_HOLD_SECS = "hold_secs"
ATTR_ALTERNATIVE = "alternative"
ATTR_TIMEOUT = "timeout"

DOMAIN = "remote"
SCAN_INTERVAL = timedelta(seconds=30)

ENTITY_ID_FORMAT = DOMAIN + ".{}"

MIN_TIME_BETWEEN_SCANS = timedelta(seconds=10)

SERVICE_SEND_COMMAND = "send_command"
SERVICE_LEARN_COMMAND = "learn_command"
SERVICE_DELETE_COMMAND = "delete_command"
SERVICE_SYNC = "sync"

DEFAULT_NUM_REPEATS = 1
DEFAULT_DELAY_SECS = 0.4
DEFAULT_HOLD_SECS = 0

SUPPORT_LEARN_COMMAND = 1
SUPPORT_DELETE_COMMAND = 2
SUPPORT_ACTIVITY = 4

REMOTE_SERVICE_ACTIVITY_SCHEMA = make_entity_service_schema(
    {vol.Optional(ATTR_ACTIVITY): cv.string}
)


@bind_opp
def is_on(opp: OpenPeerPowerType, entity_id: str) -> bool:
    """Return if the remote is on based on the statemachine."""
    return opp.states.is_state(entity_id, STATE_ON)


async def async_setup(opp: OpenPeerPowerType, config: ConfigType) -> bool:
    """Track states and offer events for remotes."""
    component = opp.data[DOMAIN] = EntityComponent(_LOGGER, DOMAIN, opp, SCAN_INTERVAL)
    await component.async_setup(config)

    component.async_register_entity_service(
        SERVICE_TURN_OFF, REMOTE_SERVICE_ACTIVITY_SCHEMA, "async_turn_off"
    )

    component.async_register_entity_service(
        SERVICE_TURN_ON, REMOTE_SERVICE_ACTIVITY_SCHEMA, "async_turn_on"
    )

    component.async_register_entity_service(
        SERVICE_TOGGLE, REMOTE_SERVICE_ACTIVITY_SCHEMA, "async_toggle"
    )

    component.async_register_entity_service(
        SERVICE_SEND_COMMAND,
        {
            vol.Required(ATTR_COMMAND): vol.All(cv.ensure_list, [cv.string]),
            vol.Optional(ATTR_DEVICE): cv.string,
            vol.Optional(
                ATTR_NUM_REPEATS, default=DEFAULT_NUM_REPEATS
            ): cv.positive_int,
            vol.Optional(ATTR_DELAY_SECS): vol.Coerce(float),
            vol.Optional(ATTR_HOLD_SECS, default=DEFAULT_HOLD_SECS): vol.Coerce(float),
        },
        "async_send_command",
    )

    component.async_register_entity_service(
        SERVICE_LEARN_COMMAND,
        {
            vol.Optional(ATTR_DEVICE): cv.string,
            vol.Optional(ATTR_COMMAND): vol.All(cv.ensure_list, [cv.string]),
            vol.Optional(ATTR_COMMAND_TYPE): cv.string,
            vol.Optional(ATTR_ALTERNATIVE): cv.boolean,
            vol.Optional(ATTR_TIMEOUT): cv.positive_int,
        },
        "async_learn_command",
    )

    component.async_register_entity_service(
        SERVICE_DELETE_COMMAND,
        {
            vol.Required(ATTR_COMMAND): vol.All(cv.ensure_list, [cv.string]),
            vol.Optional(ATTR_DEVICE): cv.string,
        },
        "async_delete_command",
    )

    return True


async def async_setup_entry(opp: OpenPeerPowerType, entry: ConfigEntry) -> bool:
    """Set up a config entry."""
    return await cast(EntityComponent, opp.data[DOMAIN]).async_setup_entry(entry)


async def async_unload_entry(opp: OpenPeerPowerType, entry: ConfigEntry) -> bool:
    """Unload a config entry."""
    return await cast(EntityComponent, opp.data[DOMAIN]).async_unload_entry(entry)


class RemoteEntity(ToggleEntity):
    """Representation of a remote."""

    @property
    def supported_features(self) -> int:
        """Flag supported features."""
        return 0

    @property
    def current_activity(self) -> Optional[str]:
        """Active activity."""
        return None

    @property
    def activity_list(self) -> Optional[List[str]]:
        """List of available activities."""
        return None

    @property
    def state_attributes(self) -> Optional[Dict[str, Any]]:
        """Return optional state attributes."""
        if not self.supported_features & SUPPORT_ACTIVITY:
            return None

        return {
            ATTR_ACTIVITY_LIST: self.activity_list,
            ATTR_CURRENT_ACTIVITY: self.current_activity,
        }

    def send_command(self, command: Iterable[str], **kwargs: Any) -> None:
        """Send commands to a device."""
        raise NotImplementedError()

    async def async_send_command(self, command: Iterable[str], **kwargs: Any) -> None:
        """Send commands to a device."""
        assert self.opp is not None
        await self.opp.async_add_executor_job(
            ft.partial(self.send_command, command, **kwargs)
        )

    def learn_command(self, **kwargs: Any) -> None:
        """Learn a command from a device."""
        raise NotImplementedError()

    async def async_learn_command(self, **kwargs: Any) -> None:
        """Learn a command from a device."""
        assert self.opp is not None
        await self.opp.async_add_executor_job(ft.partial(self.learn_command, **kwargs))

    def delete_command(self, **kwargs: Any) -> None:
        """Delete commands from the database."""
        raise NotImplementedError()

    async def async_delete_command(self, **kwargs: Any) -> None:
        """Delete commands from the database."""
        assert self.opp is not None
        await self.opp.async_add_executor_job(ft.partial(self.delete_command, **kwargs))


class RemoteDevice(RemoteEntity):
    """Representation of a remote (for backwards compatibility)."""

    def __init_subclass__(cls, **kwargs):
        """Print deprecation warning."""
        super().__init_subclass__(**kwargs)
        _LOGGER.warning(
            "RemoteDevice is deprecated, modify %s to extend RemoteEntity",
            cls.__name__,
        )



"""Support for w800rf32 devices."""
import logging

import W800rf32 as w800
import voluptuous as vol

from openpeerpower.const import (
    CONF_DEVICE,
    EVENT_OPENPEERPOWER_START,
    EVENT_OPENPEERPOWER_STOP,
)
import openpeerpower.helpers.config_validation as cv
from openpeerpower.helpers.dispatcher import dispatcher_send

DATA_W800RF32 = "data_w800rf32"
DOMAIN = "w800rf32"

W800RF32_DEVICE = "w800rf32_{}"

_LOGGER = logging.getLogger(__name__)

CONFIG_SCHEMA = vol.Schema(
    {DOMAIN: vol.Schema({vol.Required(CONF_DEVICE): cv.string})}, extra=vol.ALLOW_EXTRA
)


def setup(opp, config):
    """Set up the w800rf32 component."""

    # Declare the Handle event
    def handle_receive(event):
        """Handle received messages from w800rf32 gateway."""
        # Log event
        if not event.device:
            return
        _LOGGER.debug("Receive W800rf32 event in handle_receive")

        # Get device_type from device_id in opp.data
        device_id = event.device.lower()
        signal = W800RF32_DEVICE.format(device_id)
        dispatcher_send(opp, signal, event)

    # device --> /dev/ttyUSB0
    device = config[DOMAIN][CONF_DEVICE]
    w800_object = w800.Connect(device, None)

    def _start_w800rf32(event):
        w800_object.event_callback = handle_receive

    opp.bus.listen_once(EVENT_OPENPEERPOWER_START, _start_w800rf32)

    def _shutdown_w800rf32(event):
        """Close connection with w800rf32."""
        w800_object.close_connection()

    opp.bus.listen_once(EVENT_OPENPEERPOWER_STOP, _shutdown_w800rf32)

    opp.data[DATA_W800RF32] = w800_object

    return True



"""Offer Open Peer Power core automation rules."""
import voluptuous as vol

from openpeerpower.const import CONF_EVENT, CONF_PLATFORM, EVENT_OPENPEERPOWER_STOP
from openpeerpower.core import OppJob, callback

# mypy: allow-untyped-defs

EVENT_START = "start"
EVENT_SHUTDOWN = "shutdown"

TRIGGER_SCHEMA = vol.Schema(
    {
        vol.Required(CONF_PLATFORM): "openpeerpower",
        vol.Required(CONF_EVENT): vol.Any(EVENT_START, EVENT_SHUTDOWN),
    }
)


async def async_attach_trigger(opp, config, action, automation_info):
    """Listen for events based on configuration."""
    event = config.get(CONF_EVENT)
    job = OppJob(action)

    if event == EVENT_SHUTDOWN:

        @callback
        def opp_shutdown(event):
            """Execute when Open Peer Power is shutting down."""
            opp.async_run_opp_job(
                job,
                {
                    "trigger": {
                        "platform": "openpeerpower",
                        "event": event,
                        "description": "Open Peer Power stopping",
                    }
                },
                event.context,
            )

        return opp.bus.async_listen_once(EVENT_OPENPEERPOWER_STOP, opp_shutdown)

    # Automation are enabled while opp is starting up, fire right away
    # Check state because a config reload shouldn't trigger it.
    if automation_info["open_peer_power_start"]:
        opp.async_run_opp_job(
            job,
            {
                "trigger": {
                    "platform": "openpeerpower",
                    "event": event,
                    "description": "Open Peer Power starting",
                }
            },
        )

    return lambda: None



"""API for Netatmo bound to OPP OAuth."""
from asyncio import run_coroutine_threadsafe

import pyatmo

from openpeerpower import config_entries, core
from openpeerpower.helpers import config_entry_oauth2_flow


class ConfigEntryNetatmoAuth(pyatmo.auth.NetatmoOAuth2):
    """Provide Netatmo authentication tied to an OAuth2 based config entry."""

    def __init__(
        self,
        opp: core.OpenPeerPower,
        config_entry: config_entries.ConfigEntry,
        implementation: config_entry_oauth2_flow.AbstractOAuth2Implementation,
    ):
        """Initialize Netatmo Auth."""
        self.opp = opp
        self.session = config_entry_oauth2_flow.OAuth2Session(
            opp, config_entry, implementation
        )
        super().__init__(token=self.session.token)

    def refresh_tokens(
        self,
    ) -> dict:
        """Refresh and return new Netatmo tokens using Open Peer Power OAuth2 session."""
        run_coroutine_threadsafe(
            self.session.async_ensure_token_valid(), self.opp.loop
        ).result()

        return self.session.token



"""Sensor platform for Opp.io addons."""
from typing import Callable, List

from openpeerpower.config_entries import ConfigEntry
from openpeerpower.core import OpenPeerPower
from openpeerpower.helpers.entity import Entity

from . import ADDONS_COORDINATOR
from .const import ATTR_VERSION, ATTR_VERSION_LATEST
from .entity import OppioAddonEntity, OppioOSEntity


async def async_setup_entry(
    opp: OpenPeerPower,
    config_entry: ConfigEntry,
    async_add_entities: Callable[[List[Entity], bool], None],
) -> None:
    """Sensor set up for Opp.io config entry."""
    coordinator = opp.data[ADDONS_COORDINATOR]

    entities = []

    for attribute_name, sensor_name in (
        (ATTR_VERSION, "Version"),
        (ATTR_VERSION_LATEST, "Newest Version"),
    ):
        for addon in coordinator.data["addons"].values():
            entities.append(
                OppioAddonSensor(coordinator, addon, attribute_name, sensor_name)
            )
        if coordinator.is_opp_os:
            entities.append(OppioOSSensor(coordinator, attribute_name, sensor_name))

    async_add_entities(entities)


class OppioAddonSensor(OppioAddonEntity):
    """Sensor to track a Opp.io add-on attribute."""

    @property
    def state(self) -> str:
        """Return state of entity."""
        return self.addon_info[self.attribute_name]


class OppioOSSensor(OppioOSEntity):
    """Sensor to track a Opp.io add-on attribute."""

    @property
    def state(self) -> str:
        """Return state of entity."""
        return self.os_info[self.attribute_name]



"""Solar-Log integration."""
from openpeerpower.config_entries import ConfigEntry
from openpeerpower.helpers.typing import OpenPeerPowerType


async def async_setup(opp, config):
    """Component setup, do nothing."""
    return True


async def async_setup_entry(opp: OpenPeerPowerType, entry: ConfigEntry):
    """Set up a config entry for solarlog."""
    opp.async_create_task(opp.config_entries.async_forward_entry_setup(entry, "sensor"))
    return True


async def async_unload_entry(opp, entry):
    """Unload a config entry."""
    return await opp.config_entries.async_forward_entry_unload(entry, "sensor")



"""Helper to test significant Binary Sensor state changes."""
from __future__ import annotations

from typing import Any

from openpeerpower.core import OpenPeerPower, callback


@callback
def async_check_significant_change(
    opp: OpenPeerPower,
    old_state: str,
    old_attrs: dict,
    new_state: str,
    new_attrs: dict,
    **kwargs: Any,
) -> bool | None:
    """Test if state significantly changed."""
    if old_state != new_state:
        return True

    return False



"""Errors for the Mikrotik component."""
from openpeerpower.exceptions import OpenPeerPowerError


class CannotConnect(OpenPeerPowerError):
    """Unable to connect to the hub."""


class LoginError(OpenPeerPowerError):
    """Component got logged out."""



"""Handle the frontend for Open Peer Power."""
import json
import logging
import mimetypes
import os
import pathlib
from typing import Any, Dict, Optional, Set, Tuple

from aiohttp import hdrs, web, web_urldispatcher
import jinja2
import voluptuous as vol
from yarl import URL

from openpeerpower.components import websocket_api
from openpeerpower.components.http.view import OpenPeerPowerView
from openpeerpower.config import async_opp_config_yaml
from openpeerpower.const import CONF_MODE, CONF_NAME, EVENT_THEMES_UPDATED
from openpeerpower.core import callback
from openpeerpower.helpers import service
import openpeerpower.helpers.config_validation as cv
from openpeerpower.helpers.translation import async_get_translations
from openpeerpower.loader import async_get_integration, bind_opp

from .storage import async_setup_frontend_storage

# mypy: allow-untyped-defs, no-check-untyped-defs

# Fix mimetypes for borked Windows machines
# https://github.com/openpeerpower/frontend/issues/3336
mimetypes.add_type("text/css", ".css")
mimetypes.add_type("application/javascript", ".js")


DOMAIN = "frontend"
CONF_THEMES = "themes"
CONF_EXTRA_HTML_URL = "extra_html_url"
CONF_EXTRA_HTML_URL_ES5 = "extra_html_url_es5"
CONF_EXTRA_MODULE_URL = "extra_module_url"
CONF_EXTRA_JS_URL_ES5 = "extra_js_url_es5"
CONF_FRONTEND_REPO = "development_repo"
CONF_JS_VERSION = "javascript_version"
EVENT_PANELS_UPDATED = "panels_updated"

DEFAULT_THEME_COLOR = "#03A9F4"

MANIFEST_JSON = {
    "background_color": "#FFFFFF",
    "description": "Home automation platform that puts local control and privacy first.",
    "dir": "ltr",
    "display": "standalone",
    "icons": [
        {
            "src": f"/static/icons/favicon-{size}x{size}.png",
            "sizes": f"{size}x{size}",
            "type": "image/png",
            "purpose": "maskable any",
        }
        for size in (192, 384, 512, 1024)
    ],
    "lang": "en-US",
    "name": "Open Peer Power",
    "short_name": "OPP",
    "start_url": "/?homescreen=1",
    "theme_color": DEFAULT_THEME_COLOR,
    "prefer_related_applications": True,
    "related_applications": [
        {"platform": "play", "id": "io.openpeerpower.companion.android"}
    ],
}

DATA_PANELS = "frontend_panels"
DATA_JS_VERSION = "frontend_js_version"
DATA_EXTRA_MODULE_URL = "frontend_extra_module_url"
DATA_EXTRA_JS_URL_ES5 = "frontend_extra_js_url_es5"

THEMES_STORAGE_KEY = f"{DOMAIN}_theme"
THEMES_STORAGE_VERSION = 1
THEMES_SAVE_DELAY = 60
DATA_THEMES_STORE = "frontend_themes_store"
DATA_THEMES = "frontend_themes"
DATA_DEFAULT_THEME = "frontend_default_theme"
DATA_DEFAULT_DARK_THEME = "frontend_default_dark_theme"
DEFAULT_THEME = "default"
VALUE_NO_THEME = "none"

PRIMARY_COLOR = "primary-color"

_LOGGER = logging.getLogger(__name__)

CONFIG_SCHEMA = vol.Schema(
    {
        DOMAIN: vol.Schema(
            {
                vol.Optional(CONF_FRONTEND_REPO): cv.isdir,
                vol.Optional(CONF_THEMES): vol.Schema(
                    {cv.string: {cv.string: cv.string}}
                ),
                vol.Optional(CONF_EXTRA_MODULE_URL): vol.All(
                    cv.ensure_list, [cv.string]
                ),
                vol.Optional(CONF_EXTRA_JS_URL_ES5): vol.All(
                    cv.ensure_list, [cv.string]
                ),
                # We no longer use these options.
                vol.Optional(CONF_EXTRA_HTML_URL): cv.match_all,
                vol.Optional(CONF_EXTRA_HTML_URL_ES5): cv.match_all,
                vol.Optional(CONF_JS_VERSION): cv.match_all,
            },
        )
    },
    extra=vol.ALLOW_EXTRA,
)

SERVICE_SET_THEME = "set_theme"
SERVICE_RELOAD_THEMES = "reload_themes"


class Panel:
    """Abstract class for panels."""

    # Name of the webcomponent
    component_name: Optional[str] = None

    # Icon to show in the sidebar
    sidebar_icon: Optional[str] = None

    # Title to show in the sidebar
    sidebar_title: Optional[str] = None

    # Url to show the panel in the frontend
    frontend_url_path: Optional[str] = None

    # Config to pass to the webcomponent
    config: Optional[Dict[str, Any]] = None

    # If the panel should only be visible to admins
    require_admin = False

    def __init__(
        self,
        component_name,
        sidebar_title,
        sidebar_icon,
        frontend_url_path,
        config,
        require_admin,
    ):
        """Initialize a built-in panel."""
        self.component_name = component_name
        self.sidebar_title = sidebar_title
        self.sidebar_icon = sidebar_icon
        self.frontend_url_path = frontend_url_path or component_name
        self.config = config
        self.require_admin = require_admin

    @callback
    def to_response(self):
        """Panel as dictionary."""
        return {
            "component_name": self.component_name,
            "icon": self.sidebar_icon,
            "title": self.sidebar_title,
            "config": self.config,
            "url_path": self.frontend_url_path,
            "require_admin": self.require_admin,
        }


@bind_opp
@callback
def async_register_built_in_panel(
    opp,
    component_name,
    sidebar_title=None,
    sidebar_icon=None,
    frontend_url_path=None,
    config=None,
    require_admin=False,
    *,
    update=False,
):
    """Register a built-in panel."""
    panel = Panel(
        component_name,
        sidebar_title,
        sidebar_icon,
        frontend_url_path,
        config,
        require_admin,
    )

    panels = opp.data.setdefault(DATA_PANELS, {})

    if not update and panel.frontend_url_path in panels:
        raise ValueError(f"Overwriting panel {panel.frontend_url_path}")

    panels[panel.frontend_url_path] = panel

    opp.bus.async_fire(EVENT_PANELS_UPDATED)


@bind_opp
@callback
def async_remove_panel(opp, frontend_url_path):
    """Remove a built-in panel."""
    panel = opp.data.get(DATA_PANELS, {}).pop(frontend_url_path, None)

    if panel is None:
        _LOGGER.warning("Removing unknown panel %s", frontend_url_path)

    opp.bus.async_fire(EVENT_PANELS_UPDATED)


def add_extra_js_url(opp, url, es5=False):
    """Register extra js or module url to load."""
    key = DATA_EXTRA_JS_URL_ES5 if es5 else DATA_EXTRA_MODULE_URL
    url_set = opp.data.get(key)
    if url_set is None:
        url_set = opp.data[key] = set()
    url_set.add(url)


def add_manifest_json_key(key, val):
    """Add a keyval to the manifest.json."""
    MANIFEST_JSON[key] = val


def _frontend_root(dev_repo_path):
    """Return root path to the frontend files."""
    if dev_repo_path is not None:
        return pathlib.Path(dev_repo_path) / "opp_frontend"
    # Keep import here so that we can import frontend without installing reqs
    # pylint: disable=import-outside-toplevel
    import opp_frontend

    return opp_frontend.where()


async def async_setup(opp, config):
    """Set up the serving of the frontend."""
    await async_setup_frontend_storage(opp)
    opp.components.websocket_api.async_register_command(websocket_get_panels)
    opp.components.websocket_api.async_register_command(websocket_get_themes)
    opp.components.websocket_api.async_register_command(websocket_get_translations)
    opp.components.websocket_api.async_register_command(websocket_get_version)
    opp.http.register_view(ManifestJSONView)

    conf = config.get(DOMAIN, {})

    for key in (CONF_EXTRA_HTML_URL, CONF_EXTRA_HTML_URL_ES5, CONF_JS_VERSION):
        if key in conf:
            _LOGGER.error(
                "Please remove %s from your frontend config. It is no longer supported",
                key,
            )

    repo_path = conf.get(CONF_FRONTEND_REPO)
    is_dev = repo_path is not None
    root_path = _frontend_root(repo_path)

    for path, should_cache in (
        ("service_worker.js", False),
        ("robots.txt", False),
        ("onboarding.html", not is_dev),
        ("static", not is_dev),
        ("frontend_latest", not is_dev),
        ("frontend_es5", not is_dev),
    ):
        opp.http.register_static_path(f"/{path}", str(root_path / path), should_cache)

    opp.http.register_static_path(
        "/auth/authorize", str(root_path / "authorize.html"), False
    )
    # https://wicg.github.io/change-password-url/
    opp.http.register_redirect(
        "/.well-known/change-password", "/profile", redirect_exc=web.HTTPFound
    )

    local = opp.config.path("www")
    if os.path.isdir(local):
        opp.http.register_static_path("/local", local, not is_dev)

    opp.http.app.router.register_resource(IndexView(repo_path, opp))

    async_register_built_in_panel(opp, "profile")

    # To smooth transition to new urls, add redirects to new urls of dev tools
    # Added June 27, 2019. Can be removed in 2021.
    for panel in ("event", "service", "state", "template"):
        opp.http.register_redirect(f"/dev-{panel}", f"/developer-tools/{panel}")
    for panel in ("logs", "info", "mqtt"):
        # Can be removed in 2021.
        opp.http.register_redirect(f"/dev-{panel}", f"/config/{panel}")
        # Added June 20 2020. Can be removed in 2022.
        opp.http.register_redirect(f"/developer-tools/{panel}", f"/config/{panel}")

    async_register_built_in_panel(
        opp,
        "developer-tools",
        require_admin=True,
        sidebar_title="developer_tools",
        sidebar_icon="opp:hammer",
    )

    if DATA_EXTRA_MODULE_URL not in opp.data:
        opp.data[DATA_EXTRA_MODULE_URL] = set()

    for url in conf.get(CONF_EXTRA_MODULE_URL, []):
        add_extra_js_url(opp, url)

    if DATA_EXTRA_JS_URL_ES5 not in opp.data:
        opp.data[DATA_EXTRA_JS_URL_ES5] = set()

    for url in conf.get(CONF_EXTRA_JS_URL_ES5, []):
        add_extra_js_url(opp, url, True)

    await _async_setup_themes(opp, conf.get(CONF_THEMES))

    return True


async def _async_setup_themes(opp, themes):
    """Set up themes data and services."""
    opp.data[DATA_THEMES] = themes or {}

    store = opp.data[DATA_THEMES_STORE] = opp.helpers.storage.Store(
        THEMES_STORAGE_VERSION, THEMES_STORAGE_KEY
    )

    theme_data = await store.async_load() or {}
    theme_name = theme_data.get(DATA_DEFAULT_THEME, DEFAULT_THEME)
    dark_theme_name = theme_data.get(DATA_DEFAULT_DARK_THEME)

    if theme_name == DEFAULT_THEME or theme_name in opp.data[DATA_THEMES]:
        opp.data[DATA_DEFAULT_THEME] = theme_name
    else:
        opp.data[DATA_DEFAULT_THEME] = DEFAULT_THEME

    if dark_theme_name == DEFAULT_THEME or dark_theme_name in opp.data[DATA_THEMES]:
        opp.data[DATA_DEFAULT_DARK_THEME] = dark_theme_name

    @callback
    def update_theme_and_fire_event():
        """Update theme_color in manifest."""
        name = opp.data[DATA_DEFAULT_THEME]
        themes = opp.data[DATA_THEMES]
        MANIFEST_JSON["theme_color"] = DEFAULT_THEME_COLOR
        if name != DEFAULT_THEME:
            MANIFEST_JSON["theme_color"] = themes[name].get(
                "app-header-background-color",
                themes[name].get(PRIMARY_COLOR, DEFAULT_THEME_COLOR),
            )
        opp.bus.async_fire(EVENT_THEMES_UPDATED)

    @callback
    def set_theme(call):
        """Set backend-preferred theme."""
        name = call.data[CONF_NAME]
        mode = call.data.get("mode", "light")

        if (
            name not in (DEFAULT_THEME, VALUE_NO_THEME)
            and name not in opp.data[DATA_THEMES]
        ):
            _LOGGER.warning("Theme %s not found", name)
            return

        light_mode = mode == "light"

        theme_key = DATA_DEFAULT_THEME if light_mode else DATA_DEFAULT_DARK_THEME

        if name == VALUE_NO_THEME:
            to_set = DEFAULT_THEME if light_mode else None
        else:
            _LOGGER.info("Theme %s set as default %s theme", name, mode)
            to_set = name

        opp.data[theme_key] = to_set
        store.async_delay_save(
            lambda: {
                DATA_DEFAULT_THEME: opp.data[DATA_DEFAULT_THEME],
                DATA_DEFAULT_DARK_THEME: opp.data.get(DATA_DEFAULT_DARK_THEME),
            },
            THEMES_SAVE_DELAY,
        )
        update_theme_and_fire_event()

    async def reload_themes(_):
        """Reload themes."""
        config = await async_opp_config_yaml(opp)
        new_themes = config[DOMAIN].get(CONF_THEMES, {})
        opp.data[DATA_THEMES] = new_themes
        if opp.data[DATA_DEFAULT_THEME] not in new_themes:
            opp.data[DATA_DEFAULT_THEME] = DEFAULT_THEME
        if (
            opp.data.get(DATA_DEFAULT_DARK_THEME)
            and opp.data.get(DATA_DEFAULT_DARK_THEME) not in new_themes
        ):
            opp.data[DATA_DEFAULT_DARK_THEME] = None
        update_theme_and_fire_event()

    service.async_register_admin_service(
        opp,
        DOMAIN,
        SERVICE_SET_THEME,
        set_theme,
        vol.Schema(
            {
                vol.Required(CONF_NAME): cv.string,
                vol.Optional(CONF_MODE): vol.Any("dark", "light"),
            }
        ),
    )

    service.async_register_admin_service(
        opp, DOMAIN, SERVICE_RELOAD_THEMES, reload_themes
    )


class IndexView(web_urldispatcher.AbstractResource):
    """Serve the frontend."""

    def __init__(self, repo_path, opp):
        """Initialize the frontend view."""
        super().__init__(name="frontend:index")
        self.repo_path = repo_path
        self.opp = opp
        self._template_cache = None

    @property
    def canonical(self) -> str:
        """Return resource's canonical path."""
        return "/"

    @property
    def _route(self):
        """Return the index route."""
        return web_urldispatcher.ResourceRoute("GET", self.get, self)

    def url_for(self, **kwargs: str) -> URL:
        """Construct url for resource with additional params."""
        return URL("/")

    async def resolve(
        self, request: web.Request
    ) -> Tuple[Optional[web_urldispatcher.UrlMappingMatchInfo], Set[str]]:
        """Resolve resource.

        Return (UrlMappingMatchInfo, allowed_methods) pair.
        """
        if (
            request.path != "/"
            and request.url.parts[1] not in self.opp.data[DATA_PANELS]
        ):
            return None, set()

        if request.method != hdrs.METH_GET:
            return None, {"GET"}

        return web_urldispatcher.UrlMappingMatchInfo({}, self._route), {"GET"}

    def add_prefix(self, prefix: str) -> None:
        """Add a prefix to processed URLs.

        Required for subapplications support.
        """

    def get_info(self):
        """Return a dict with additional info useful for introspection."""
        return {"panels": list(self.opp.data[DATA_PANELS])}

    def freeze(self) -> None:
        """Freeze the resource."""

    def raw_match(self, path: str) -> bool:
        """Perform a raw match against path."""

    def get_template(self):
        """Get template."""
        tpl = self._template_cache
        if tpl is None:
            with open(str(_frontend_root(self.repo_path) / "index.html")) as file:
                tpl = jinja2.Template(file.read())

            # Cache template if not running from repository
            if self.repo_path is None:
                self._template_cache = tpl

        return tpl

    async def get(self, request: web.Request) -> web.Response:
        """Serve the index page for panel pages."""
        opp = request.app["opp"]

        if not opp.components.onboarding.async_is_onboarded():
            return web.Response(status=302, headers={"location": "/onboarding.html"})

        template = self._template_cache

        if template is None:
            template = await opp.async_add_executor_job(self.get_template)

        return web.Response(
            text=template.render(
                theme_color=MANIFEST_JSON["theme_color"],
                extra_modules=opp.data[DATA_EXTRA_MODULE_URL],
                extra_js_es5=opp.data[DATA_EXTRA_JS_URL_ES5],
            ),
            content_type="text/html",
        )

    def __len__(self) -> int:
        """Return length of resource."""
        return 1

    def __iter__(self):
        """Iterate over routes."""
        return iter([self._route])


class ManifestJSONView(OpenPeerPowerView):
    """View to return a manifest.json."""

    requires_auth = False
    url = "/manifest.json"
    name = "manifestjson"

    @callback
    def get(self, request):  # pylint: disable=no-self-use
        """Return the manifest.json."""
        msg = json.dumps(MANIFEST_JSON, sort_keys=True)
        return web.Response(text=msg, content_type="application/manifest+json")


@callback
@websocket_api.websocket_command({"type": "get_panels"})
def websocket_get_panels(opp, connection, msg):
    """Handle get panels command."""
    user_is_admin = connection.user.is_admin
    panels = {
        panel_key: panel.to_response()
        for panel_key, panel in connection.opp.data[DATA_PANELS].items()
        if user_is_admin or not panel.require_admin
    }

    connection.send_message(websocket_api.result_message(msg["id"], panels))


@callback
@websocket_api.websocket_command({"type": "frontend/get_themes"})
def websocket_get_themes(opp, connection, msg):
    """Handle get themes command."""
    if opp.config.safe_mode:
        connection.send_message(
            websocket_api.result_message(
                msg["id"],
                {
                    "themes": {
                        "safe_mode": {
                            "primary-color": "#db4437",
                            "accent-color": "#ffca28",
                        }
                    },
                    "default_theme": "safe_mode",
                },
            )
        )
        return

    connection.send_message(
        websocket_api.result_message(
            msg["id"],
            {
                "themes": opp.data[DATA_THEMES],
                "default_theme": opp.data[DATA_DEFAULT_THEME],
                "default_dark_theme": opp.data.get(DATA_DEFAULT_DARK_THEME),
            },
        )
    )


@websocket_api.websocket_command(
    {
        "type": "frontend/get_translations",
        vol.Required("language"): str,
        vol.Required("category"): str,
        vol.Optional("integration"): str,
        vol.Optional("config_flow"): bool,
    }
)
@websocket_api.async_response
async def websocket_get_translations(opp, connection, msg):
    """Handle get translations command."""
    resources = await async_get_translations(
        opp,
        msg["language"],
        msg["category"],
        msg.get("integration"),
        msg.get("config_flow"),
    )
    connection.send_message(
        websocket_api.result_message(msg["id"], {"resources": resources})
    )


@websocket_api.websocket_command({"type": "frontend/get_version"})
@websocket_api.async_response
async def websocket_get_version(opp, connection, msg):
    """Handle get version command."""
    integration = await async_get_integration(opp, "frontend")

    frontend = None

    for req in integration.requirements:
        if req.startswith("open-peer-power-frontend=="):
            frontend = req.split("==", 1)[1]

    if frontend is None:
        connection.send_error(msg["id"], "unknown_version", "Version not found")
    else:
        connection.send_result(msg["id"], {"version": frontend})



"""Provide configuration end points for Customize."""
from openpeerpower.components.openpeerpower import SERVICE_RELOAD_CORE_CONFIG
from openpeerpower.config import DATA_CUSTOMIZE
from openpeerpower.core import DOMAIN
import openpeerpower.helpers.config_validation as cv

from . import EditKeyBasedConfigView

CONFIG_PATH = "customize.yaml"


async def async_setup(opp):
    """Set up the Customize config API."""

    async def hook(action, config_key):
        """post_write_hook for Config View that reloads groups."""
        await opp.services.async_call(DOMAIN, SERVICE_RELOAD_CORE_CONFIG)

    opp.http.register_view(
        CustomizeConfigView(
            "customize", "config", CONFIG_PATH, cv.entity_id, dict, post_write_hook=hook
        )
    )

    return True


class CustomizeConfigView(EditKeyBasedConfigView):
    """Configure a list of entries."""

    def _get_value(self, opp, data, config_key):
        """Get value."""
        customize = opp.data.get(DATA_CUSTOMIZE, {}).get(config_key) or {}
        return {"global": customize, "local": data.get(config_key, {})}

    def _write_value(self, opp, data, config_key, new_value):
        """Set value."""
        data[config_key] = new_value

        state = opp.states.get(config_key)
        state_attributes = dict(state.attributes)
        state_attributes.update(new_value)
        opp.states.async_set(config_key, state.state, state_attributes)



"""Zerproc lights integration."""
import asyncio

from openpeerpower.config_entries import SOURCE_IMPORT, ConfigEntry
from openpeerpower.core import OpenPeerPower

from .const import DOMAIN

PLATFORMS = ["light"]


async def async_setup(opp, config):
    """Set up the Zerproc platform."""
    opp.async_create_task(
        opp.config_entries.flow.async_init(DOMAIN, context={"source": SOURCE_IMPORT})
    )

    return True


async def async_setup_entry(opp: OpenPeerPower, entry: ConfigEntry):
    """Set up Zerproc from a config entry."""
    for platform in PLATFORMS:
        opp.async_create_task(
            opp.config_entries.async_forward_entry_setup(entry, platform)
        )

    return True


async def async_unload_entry(opp: OpenPeerPower, entry: ConfigEntry):
    """Unload a config entry."""
    return all(
        await asyncio.gather(
            *[
                opp.config_entries.async_forward_entry_unload(entry, platform)
                for platform in PLATFORMS
            ]
        )
    )



"""Errors for media source."""
from openpeerpower.exceptions import OpenPeerPowerError


class MediaSourceError(OpenPeerPowerError):
    """Base class for media source errors."""


class Unresolvable(MediaSourceError):
    """When media ID is not resolvable."""



"""The twinkly component."""

import twinkly_client

from openpeerpower.config_entries import ConfigEntry
from openpeerpower.helpers.aiohttp_client import async_get_clientsession
from openpeerpower.helpers.typing import OpenPeerPowerType

from .const import CONF_ENTRY_HOST, CONF_ENTRY_ID, DOMAIN


async def async_setup(opp: OpenPeerPowerType, config: dict):
    """Set up the twinkly integration."""

    return True


async def async_setup_entry(opp: OpenPeerPowerType, config_entry: ConfigEntry):
    """Set up entries from config flow."""

    # We setup the client here so if at some point we add any other entity for this device,
    # we will be able to properly share the connection.
    uuid = config_entry.data[CONF_ENTRY_ID]
    host = config_entry.data[CONF_ENTRY_HOST]

    opp.data.setdefault(DOMAIN, {})[uuid] = twinkly_client.TwinklyClient(
        host, async_get_clientsession(opp)
    )

    opp.async_create_task(
        opp.config_entries.async_forward_entry_setup(config_entry, "light")
    )
    return True


async def async_unload_entry(opp: OpenPeerPowerType, config_entry: ConfigEntry):
    """Remove a twinkly entry."""

    # For now light entries don't have unload method, so we don't have to async_forward_entry_unload
    # However we still have to cleanup the shared client!
    uuid = config_entry.data[CONF_ENTRY_ID]
    opp.data[DOMAIN].pop(uuid)

    return True



"""Support for ADS binary sensors."""
import voluptuous as vol

from openpeerpower.components.binary_sensor import (
    DEVICE_CLASS_MOVING,
    DEVICE_CLASSES_SCHEMA,
    PLATFORM_SCHEMA,
    BinarySensorEntity,
)
from openpeerpower.const import CONF_DEVICE_CLASS, CONF_NAME
import openpeerpower.helpers.config_validation as cv

from . import CONF_ADS_VAR, DATA_ADS, STATE_KEY_STATE, AdsEntity

DEFAULT_NAME = "ADS binary sensor"
PLATFORM_SCHEMA = PLATFORM_SCHEMA.extend(
    {
        vol.Required(CONF_ADS_VAR): cv.string,
        vol.Optional(CONF_NAME, default=DEFAULT_NAME): cv.string,
        vol.Optional(CONF_DEVICE_CLASS): DEVICE_CLASSES_SCHEMA,
    }
)


def setup_platform(opp, config, add_entities, discovery_info=None):
    """Set up the Binary Sensor platform for ADS."""
    ads_hub = opp.data.get(DATA_ADS)

    ads_var = config[CONF_ADS_VAR]
    name = config[CONF_NAME]
    device_class = config.get(CONF_DEVICE_CLASS)

    ads_sensor = AdsBinarySensor(ads_hub, name, ads_var, device_class)
    add_entities([ads_sensor])


class AdsBinarySensor(AdsEntity, BinarySensorEntity):
    """Representation of ADS binary sensors."""

    def __init__(self, ads_hub, name, ads_var, device_class):
        """Initialize ADS binary sensor."""
        super().__init__(ads_hub, name, ads_var)
        self._device_class = device_class or DEVICE_CLASS_MOVING

    async def async_added_to_opp(self):
        """Register device notification."""
        await self.async_initialize_device(self._ads_var, self._ads_hub.PLCTYPE_BOOL)

    @property
    def is_on(self):
        """Return True if the entity is on."""
        return self._state_dict[STATE_KEY_STATE]

    @property
    def device_class(self):
        """Return the device class."""
        return self._device_class



"""The forked_daapd component."""
from openpeerpower.components.media_player import DOMAIN as MP_DOMAIN

from .const import DOMAIN, OPP_DATA_REMOVE_LISTENERS_KEY, OPP_DATA_UPDATER_KEY


async def async_setup(opp, config):
    """Set up the forked-daapd component."""
    return True


async def async_setup_entry(opp, entry):
    """Set up forked-daapd from a config entry by forwarding to platform."""
    opp.async_create_task(
        opp.config_entries.async_forward_entry_setup(entry, MP_DOMAIN)
    )
    return True


async def async_unload_entry(opp, entry):
    """Remove forked-daapd component."""
    status = await opp.config_entries.async_forward_entry_unload(entry, MP_DOMAIN)
    if status and opp.data.get(DOMAIN) and opp.data[DOMAIN].get(entry.entry_id):
        opp.data[DOMAIN][entry.entry_id][
            OPP_DATA_UPDATER_KEY
        ].websocket_handler.cancel()
        for remove_listener in opp.data[DOMAIN][entry.entry_id][
            OPP_DATA_REMOVE_LISTENERS_KEY
        ]:
            remove_listener()
        del opp.data[DOMAIN][entry.entry_id]
        if not opp.data[DOMAIN]:
            del opp.data[DOMAIN]
    return status



"""Errors for the Konnected component."""
from openpeerpower.exceptions import OpenPeerPowerError


class KonnectedException(OpenPeerPowerError):
    """Base class for Konnected exceptions."""


class CannotConnect(KonnectedException):
    """Unable to connect to the panel."""



"""The 1-Wire component."""
import asyncio

from openpeerpower.config_entries import ConfigEntry
from openpeerpower.exceptions import ConfigEntryNotReady
from openpeerpower.helpers.typing import OpenPeerPowerType

from .const import DOMAIN, PLATFORMS
from .onewirehub import CannotConnect, OneWireHub


async def async_setup(opp, config):
    """Set up 1-Wire integrations."""
    return True


async def async_setup_entry(opp: OpenPeerPowerType, config_entry: ConfigEntry):
    """Set up a 1-Wire proxy for a config entry."""
    opp.data.setdefault(DOMAIN, {})

    onewirehub = OneWireHub(opp)
    try:
        await onewirehub.initialize(config_entry)
    except CannotConnect as exc:
        raise ConfigEntryNotReady() from exc

    opp.data[DOMAIN][config_entry.unique_id] = onewirehub

    for platform in PLATFORMS:
        opp.async_create_task(
            opp.config_entries.async_forward_entry_setup(config_entry, platform)
        )
    return True


async def async_unload_entry(opp: OpenPeerPowerType, config_entry: ConfigEntry):
    """Unload a config entry."""
    unload_ok = all(
        await asyncio.gather(
            *[
                opp.config_entries.async_forward_entry_unload(config_entry, platform)
                for platform in PLATFORMS
            ]
        )
    )
    if unload_ok:
        opp.data[DOMAIN].pop(config_entry.unique_id)
    return unload_ok



"""The syncthru component."""

import logging
from typing import Set, Tuple

from pysyncthru import SyncThru

from openpeerpower.components.sensor import DOMAIN as SENSOR_DOMAIN
from openpeerpower.config_entries import ConfigEntry
from openpeerpower.const import CONF_URL
from openpeerpower.exceptions import ConfigEntryNotReady
from openpeerpower.helpers import aiohttp_client, device_registry as dr
from openpeerpower.helpers.typing import ConfigType, OpenPeerPowerType

from .const import DOMAIN

_LOGGER = logging.getLogger(__name__)


async def async_setup(opp: OpenPeerPowerType, config: ConfigType) -> bool:
    """Set up."""
    opp.data.setdefault(DOMAIN, {})
    return True


async def async_setup_entry(opp: OpenPeerPowerType, entry: ConfigEntry) -> bool:
    """Set up config entry."""

    session = aiohttp_client.async_get_clientsession(opp)
    printer = opp.data[DOMAIN][entry.entry_id] = SyncThru(entry.data[CONF_URL], session)

    try:
        await printer.update()
    except ValueError:
        _LOGGER.error(
            "Device at %s not appear to be a SyncThru printer, aborting setup",
            printer.url,
        )
        return False
    else:
        if printer.is_unknown_state():
            raise ConfigEntryNotReady

    device_registry = await dr.async_get_registry(opp)
    device_registry.async_get_or_create(
        config_entry_id=entry.entry_id,
        connections=device_connections(printer),
        identifiers=device_identifiers(printer),
        model=printer.model(),
        name=printer.hostname(),
    )

    opp.async_create_task(
        opp.config_entries.async_forward_entry_setup(entry, SENSOR_DOMAIN)
    )
    return True


async def async_unload_entry(opp: OpenPeerPowerType, entry: ConfigEntry) -> bool:
    """Unload the config entry."""
    await opp.config_entries.async_forward_entry_unload(entry, SENSOR_DOMAIN)
    opp.data[DOMAIN].pop(entry.entry_id, None)
    return True


def device_identifiers(printer: SyncThru) -> Set[Tuple[str, str]]:
    """Get device identifiers for device registry."""
    return {(DOMAIN, printer.serial_number())}


def device_connections(printer: SyncThru) -> Set[Tuple[str, str]]:
    """Get device connections for device registry."""
    connections = set()
    try:
        mac = printer.raw()["identity"]["mac_addr"]
        if mac:
            connections.add((dr.CONNECTION_NETWORK_MAC, mac))
    except AttributeError:
        pass
    return connections



"""Provide configuration end points for Groups."""
from openpeerpower.components.group import (
    DOMAIN,
    GROUP_SCHEMA,
    GroupIntegrationRegistry,
)
from openpeerpower.config import GROUP_CONFIG_PATH
from openpeerpower.const import SERVICE_RELOAD
from openpeerpower.core import callback
import openpeerpower.helpers.config_validation as cv
from openpeerpower.helpers.typing import OpenPeerPowerType

from . import EditKeyBasedConfigView


async def async_setup(opp):
    """Set up the Group config API."""

    async def hook(action, config_key):
        """post_write_hook for Config View that reloads groups."""
        await opp.services.async_call(DOMAIN, SERVICE_RELOAD)

    opp.http.register_view(
        EditKeyBasedConfigView(
            "group",
            "config",
            GROUP_CONFIG_PATH,
            cv.slug,
            GROUP_SCHEMA,
            post_write_hook=hook,
        )
    )
    return True


@callback
def async_describe_on_off_states(
    opp: OpenPeerPowerType, registry: GroupIntegrationRegistry
) -> None:
    """Describe group on off states."""
    return



"""The Safe Mode integration."""
from openpeerpower.components import persistent_notification
from openpeerpower.core import OpenPeerPower

DOMAIN = "safe_mode"


async def async_setup(opp: OpenPeerPower, config: dict):
    """Set up the Safe Mode component."""
    persistent_notification.async_create(
        opp,
        "Open Peer Power is running in safe mode. Check [the error log](/config/logs) to see what went wrong.",
        "Safe Mode",
    )
    return True



"""Helper for httpx."""
import sys
from typing import Any, Callable, Optional

import httpx

from openpeerpower.const import EVENT_OPENPEERPOWER_CLOSE, __version__
from openpeerpower.core import Event, callback
from openpeerpower.helpers.frame import warn_use
from openpeerpower.helpers.typing import OpenPeerPowerType
from openpeerpower.loader import bind_opp

DATA_ASYNC_CLIENT = "httpx_async_client"
DATA_ASYNC_CLIENT_NOVERIFY = "httpx_async_client_noverify"
SERVER_SOFTWARE = "OpenPeerPower/{0} httpx/{1} Python/{2[0]}.{2[1]}".format(
    __version__, httpx.__version__, sys.version_info
)
USER_AGENT = "User-Agent"


@callback
@bind_opp
def get_async_client(
    opp: OpenPeerPowerType, verify_ssl: bool = True
) -> httpx.AsyncClient:
    """Return default httpx AsyncClient.

    This method must be run in the event loop.
    """
    key = DATA_ASYNC_CLIENT if verify_ssl else DATA_ASYNC_CLIENT_NOVERIFY

    client: Optional[httpx.AsyncClient] = opp.data.get(key)

    if client is None:
        client = opp.data[key] = create_async_httpx_client(opp, verify_ssl)

    return client


@callback
def create_async_httpx_client(
    opp: OpenPeerPowerType,
    verify_ssl: bool = True,
    auto_cleanup: bool = True,
    **kwargs: Any,
) -> httpx.AsyncClient:
    """Create a new httpx.AsyncClient with kwargs, i.e. for cookies.

    If auto_cleanup is False, the client will be
    automatically closed on openpeerpower_stop.

    This method must be run in the event loop.
    """
    client = httpx.AsyncClient(
        verify=verify_ssl,
        headers={USER_AGENT: SERVER_SOFTWARE},
        **kwargs,
    )

    original_aclose = client.aclose

    client.aclose = warn_use(  # type: ignore
        client.aclose, "closes the Open Peer Power httpx client"
    )

    if auto_cleanup:
        _async_register_async_client_shutdown(opp, client, original_aclose)

    return client


@callback
def _async_register_async_client_shutdown(
    opp: OpenPeerPowerType,
    client: httpx.AsyncClient,
    original_aclose: Callable[..., Any],
) -> None:
    """Register httpx AsyncClient aclose on Open Peer Power shutdown.

    This method must be run in the event loop.
    """

    async def _async_close_client(event: Event) -> None:
        """Close httpx client."""
        await original_aclose()

    opp.bus.async_listen_once(EVENT_OPENPEERPOWER_CLOSE, _async_close_client)



"""Config flow for Cast."""
import voluptuous as vol

from openpeerpower import config_entries
from openpeerpower.helpers import config_validation as cv

from .const import CONF_KNOWN_HOSTS, DOMAIN

KNOWN_HOSTS_SCHEMA = vol.Schema(vol.All(cv.ensure_list, [cv.string]))


class FlowHandler(config_entries.ConfigFlow, domain=DOMAIN):
    """Handle a config flow."""

    VERSION = 1
    CONNECTION_CLASS = config_entries.CONN_CLASS_LOCAL_PUSH

    def __init__(self):
        """Initialize flow."""
        self._known_hosts = None

    @staticmethod
    def async_get_options_flow(config_entry):
        """Get the options flow for this handler."""
        return CastOptionsFlowHandler(config_entry)

    async def async_step_import(self, import_data=None):
        """Import data."""
        if self._async_current_entries():
            return self.async_abort(reason="single_instance_allowed")
        data = {CONF_KNOWN_HOSTS: self._known_hosts}
        return self.async_create_entry(title="Google Cast", data=data)

    async def async_step_user(self, user_input=None):
        """Handle a flow initialized by the user."""
        if self._async_current_entries():
            return self.async_abort(reason="single_instance_allowed")

        return await self.async_step_config()

    async def async_step_zeroconf(self, discovery_info):
        """Handle a flow initialized by zeroconf discovery."""
        if self._async_in_progress() or self._async_current_entries():
            return self.async_abort(reason="single_instance_allowed")

        await self.async_set_unique_id(DOMAIN)

        return await self.async_step_confirm()

    async def async_step_config(self, user_input=None):
        """Confirm the setup."""
        errors = {}
        data = {CONF_KNOWN_HOSTS: self._known_hosts}

        if user_input is not None:
            bad_hosts = False
            known_hosts = user_input[CONF_KNOWN_HOSTS]
            known_hosts = [x.strip() for x in known_hosts.split(",") if x.strip()]
            try:
                known_hosts = KNOWN_HOSTS_SCHEMA(known_hosts)
            except vol.Invalid:
                errors["base"] = "invalid_known_hosts"
                bad_hosts = True
            else:
                data[CONF_KNOWN_HOSTS] = known_hosts
            if not bad_hosts:
                return self.async_create_entry(title="Google Cast", data=data)

        fields = {}
        fields[vol.Optional(CONF_KNOWN_HOSTS, default="")] = str

        return self.async_show_form(
            step_id="config", data_schema=vol.Schema(fields), errors=errors
        )

    async def async_step_confirm(self, user_input=None):
        """Confirm the setup."""

        data = {CONF_KNOWN_HOSTS: self._known_hosts}

        if user_input is not None:
            return self.async_create_entry(title="Google Cast", data=data)

        return self.async_show_form(step_id="confirm")


class CastOptionsFlowHandler(config_entries.OptionsFlow):
    """Handle Google Cast options."""

    def __init__(self, config_entry):
        """Initialize MQTT options flow."""
        self.config_entry = config_entry
        self.broker_config = {}
        self.options = dict(config_entry.options)

    async def async_step_init(self, user_input=None):
        """Manage the Cast options."""
        return await self.async_step_options()

    async def async_step_options(self, user_input=None):
        """Manage the MQTT options."""
        errors = {}
        current_config = self.config_entry.data
        if user_input is not None:
            bad_hosts = False

            known_hosts = user_input.get(CONF_KNOWN_HOSTS, "")
            known_hosts = [x.strip() for x in known_hosts.split(",") if x.strip()]
            try:
                known_hosts = KNOWN_HOSTS_SCHEMA(known_hosts)
            except vol.Invalid:
                errors["base"] = "invalid_known_hosts"
                bad_hosts = True
            if not bad_hosts:
                updated_config = {}
                updated_config[CONF_KNOWN_HOSTS] = known_hosts
                self.opp.config_entries.async_update_entry(
                    self.config_entry, data=updated_config
                )
                return self.async_create_entry(title="", data=None)

        fields = {}
        known_hosts_string = ""
        if current_config.get(CONF_KNOWN_HOSTS):
            known_hosts_string = ",".join(current_config.get(CONF_KNOWN_HOSTS))
        fields[
            vol.Optional(
                "known_hosts", description={"suggested_value": known_hosts_string}
            )
        ] = str

        return self.async_show_form(
            step_id="options",
            data_schema=vol.Schema(fields),
            errors=errors,
        )



"""Errors for the Transmission component."""
from openpeerpower.exceptions import OpenPeerPowerError


class AuthenticationError(OpenPeerPowerError):
    """Wrong Username or Password."""


class CannotConnect(OpenPeerPowerError):
    """Unable to connect to client."""


class UnknownError(OpenPeerPowerError):
    """Unknown Error."""



"""Plugwise platform for Open Peer Power Core."""

from openpeerpower.config_entries import ConfigEntry
from openpeerpower.const import CONF_HOST
from openpeerpower.core import OpenPeerPower

from .gateway import async_setup_entry_gw, async_unload_entry_gw


async def async_setup(opp: OpenPeerPower, config: dict):
    """Set up the Plugwise platform."""
    return True


async def async_setup_entry(opp: OpenPeerPower, entry: ConfigEntry) -> bool:
    """Set up Plugwise components from a config entry."""
    if entry.data.get(CONF_HOST):
        return await async_setup_entry_gw(opp, entry)
    # PLACEHOLDER USB entry setup
    return False


async def async_unload_entry(opp: OpenPeerPower, entry: ConfigEntry):
    """Unload the Plugwise components."""
    if entry.data.get(CONF_HOST):
        return await async_unload_entry_gw(opp, entry)
    # PLACEHOLDER USB entry setup
    return False



"""Describe group states."""


from openpeerpower.components.group import GroupIntegrationRegistry
from openpeerpower.const import STATE_HOME, STATE_NOT_HOME
from openpeerpower.core import callback
from openpeerpower.helpers.typing import OpenPeerPowerType


@callback
def async_describe_on_off_states(
    opp: OpenPeerPowerType, registry: GroupIntegrationRegistry
) -> None:
    """Describe group on off states."""
    registry.on_off_states({STATE_HOME}, STATE_NOT_HOME)



"""Errors for the Hue component."""
from openpeerpower.exceptions import OpenPeerPowerError


class HueException(OpenPeerPowerError):
    """Base class for Hue exceptions."""


class CannotConnect(HueException):
    """Unable to connect to the bridge."""


class AuthenticationRequired(HueException):
    """Unknown error occurred."""



"""Kuler Sky lights integration."""
import asyncio

from openpeerpower.config_entries import ConfigEntry
from openpeerpower.core import OpenPeerPower

from .const import DOMAIN

PLATFORMS = ["light"]


async def async_setup(opp: OpenPeerPower, config: dict):
    """Set up the Kuler Sky component."""
    return True


async def async_setup_entry(opp: OpenPeerPower, entry: ConfigEntry):
    """Set up Kuler Sky from a config entry."""
    for platform in PLATFORMS:
        opp.async_create_task(
            opp.config_entries.async_forward_entry_setup(entry, platform)
        )

    return True


async def async_unload_entry(opp: OpenPeerPower, entry: ConfigEntry):
    """Unload a config entry."""
    unload_ok = all(
        await asyncio.gather(
            *[
                opp.config_entries.async_forward_entry_unload(entry, platform)
                for platform in PLATFORMS
            ]
        )
    )
    if unload_ok:
        opp.data[DOMAIN].pop(entry.entry_id)

    return unload_ok



"""Device automation exceptions."""
from openpeerpower.exceptions import OpenPeerPowerError


class InvalidDeviceAutomationConfig(OpenPeerPowerError):
    """When device automation config is invalid."""


class DeviceNotFound(OpenPeerPowerError):
    """When referenced device not found."""



"""Support for Freebox devices (Freebox v6 and Freebox mini 4K)."""
import asyncio
import logging

import voluptuous as vol

from openpeerpower.config_entries import SOURCE_IMPORT, ConfigEntry
from openpeerpower.const import CONF_HOST, CONF_PORT, EVENT_OPENPEERPOWER_STOP
from openpeerpower.helpers import config_validation as cv
from openpeerpower.helpers.typing import OpenPeerPowerType

from .const import DOMAIN, PLATFORMS, SERVICE_REBOOT
from .router import FreeboxRouter

_LOGGER = logging.getLogger(__name__)

FREEBOX_SCHEMA = vol.Schema(
    {vol.Required(CONF_HOST): cv.string, vol.Required(CONF_PORT): cv.port}
)

CONFIG_SCHEMA = vol.Schema(
    {DOMAIN: vol.Schema(vol.All(cv.ensure_list, [FREEBOX_SCHEMA]))},
    extra=vol.ALLOW_EXTRA,
)


async def async_setup(opp, config):
    """Set up the Freebox integration."""
    if DOMAIN in config:
        for entry_config in config[DOMAIN]:
            opp.async_create_task(
                opp.config_entries.flow.async_init(
                    DOMAIN, context={"source": SOURCE_IMPORT}, data=entry_config
                )
            )

    return True


async def async_setup_entry(opp: OpenPeerPowerType, entry: ConfigEntry):
    """Set up Freebox entry."""
    router = FreeboxRouter(opp, entry)
    await router.setup()

    opp.data.setdefault(DOMAIN, {})
    opp.data[DOMAIN][entry.unique_id] = router

    for platform in PLATFORMS:
        opp.async_create_task(
            opp.config_entries.async_forward_entry_setup(entry, platform)
        )

    # Services
    async def async_reboot(call):
        """Handle reboot service call."""
        await router.reboot()

    opp.services.async_register(DOMAIN, SERVICE_REBOOT, async_reboot)

    async def async_close_connection(event):
        """Close Freebox connection on OP Stop."""
        await router.close()

    opp.bus.async_listen_once(EVENT_OPENPEERPOWER_STOP, async_close_connection)

    return True


async def async_unload_entry(opp: OpenPeerPowerType, entry: ConfigEntry):
    """Unload a config entry."""
    unload_ok = all(
        await asyncio.gather(
            *[
                opp.config_entries.async_forward_entry_unload(entry, platform)
                for platform in PLATFORMS
            ]
        )
    )
    if unload_ok:
        router = opp.data[DOMAIN].pop(entry.unique_id)
        await router.close()
        opp.services.async_remove(DOMAIN, SERVICE_REBOOT)

    return unload_ok
